/*
This file is part of Mitsuba, a physically based rendering system.

Copyright (c) 2007-2014 by Wenzel Jakob and others.

Mitsuba is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License Version 3
as published by the Free Software Foundation.

Mitsuba is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <mitsuba/core/statistics.h>
#include <mitsuba/render/integrator.h>
#include <mitsuba/render/renderproc.h>

// NOTE(yaoyi): add external include files
#include <numeric>
#include <chrono>

using namespace nvinfer1;
using namespace std;

MTS_NAMESPACE_BEGIN
//#define nolog
//#define loadlog
// NOTE(yaoyi): these are the implemetation of tensorRT inference
// the definision are inside the header files
static int getBindingInputIndex(nvinfer1::IExecutionContext* context)
{
	return !context->getEngine().bindingIsInput(0); // 0 (false) if bindingIsInput(0), 1 (true) otherwise
}

void lf_launchInference(IExecutionContext* context, cudaStream_t stream, vector<float> const& inputTensor, vector<float>& outputTensor, void** bindings, int batchSize)
{
	int inputId = getBindingInputIndex(context);
	cudaMemcpyAsync(bindings[inputId], inputTensor.data(), inputTensor.size() * sizeof(float), cudaMemcpyHostToDevice, stream);
	context->enqueueV2(bindings, stream, nullptr);
	cudaMemcpyAsync(outputTensor.data(), bindings[1 - inputId], outputTensor.size() * sizeof(float), cudaMemcpyDeviceToHost, stream);
}
void alpha_launchInference(IExecutionContext* context, cudaStream_t stream, vector<float> const& inputTensor, vector<float>& outputTensor, void** bindings, int batchSize)
{
	int inputId = getBindingInputIndex(context);
	cudaMemcpyAsync(bindings[inputId], inputTensor.data(), inputTensor.size() * sizeof(float), cudaMemcpyHostToDevice, stream);
	context->enqueueV2(bindings, stream, nullptr);
	cudaMemcpyAsync(outputTensor.data(), bindings[1 - inputId], outputTensor.size() * sizeof(float), cudaMemcpyDeviceToHost, stream);
}
void sw_launchInference(IExecutionContext* context, cudaStream_t stream, vector<float> const& inputTensor, vector<float>& outputTensor, void** bindings, int batchSize)
{
	int inputId = getBindingInputIndex(context);
	cudaMemcpyAsync(bindings[inputId], inputTensor.data(), inputTensor.size() * sizeof(float), cudaMemcpyHostToDevice, stream);
	context->enqueueV2(bindings, stream, nullptr);
	cudaMemcpyAsync(outputTensor.data(), bindings[1 - inputId], outputTensor.size() * sizeof(float), cudaMemcpyDeviceToHost, stream);
}

Integrator::Integrator(const Properties &props)
	: NetworkedObject(props) { }

Integrator::Integrator(Stream *stream, InstanceManager *manager)
	: NetworkedObject(stream, manager) { }

bool Integrator::preprocess(const Scene *scene, RenderQueue *queue, const RenderJob *job,
	int sceneResID, int sensorResID, int samplerResID) {
	return true;
}
void Integrator::postprocess(const Scene *scene, RenderQueue *queue, const RenderJob *job,
	int sceneResID, int sensorResID, int samplerResID) { }
void Integrator::serialize(Stream *stream, InstanceManager *manager) const {
	NetworkedObject::serialize(stream, manager);
}
void Integrator::configureSampler(const Scene *scene, Sampler *sampler) {
	/* Prepare the sampler for bucket-based rendering */
	sampler->setFilmResolution(scene->getFilm()->getCropSize(),
		getClass()->derivesFrom(MTS_CLASS(SamplingIntegrator)));
}
const Integrator *Integrator::getSubIntegrator(int idx) const { return NULL; }

SamplingIntegrator::SamplingIntegrator(const Properties &props)
	: Integrator(props) { }

SamplingIntegrator::SamplingIntegrator(Stream *stream, InstanceManager *manager)
	: Integrator(stream, manager) { }

void SamplingIntegrator::serialize(Stream *stream, InstanceManager *manager) const {
	Integrator::serialize(stream, manager);
}

float clamp(float a, float b, float c)
{
	if (a < b)
		return b;
	if (a > c)
		return c;
	return a;
}
Spectrum SamplingIntegrator::E(const Scene *scene, const Intersection &its,
	const Medium *medium, Sampler *sampler, int nSamples, bool handleIndirect) {
	Spectrum E(0.0f);
	RadianceQueryRecord query(scene, sampler);
	DirectSamplingRecord dRec(its);
	Frame frame(its.shFrame.n);

	sampler->generate(Point2i(0));
	for (int i = 0; i<nSamples; i++) {
		/* Sample the direct illumination component */
		int maxIntermediateInteractions = -1;
		Spectrum directRadiance = scene->sampleAttenuatedEmitterDirect(
			dRec, its, medium, maxIntermediateInteractions, query.nextSample2D());

		if (!directRadiance.isZero()) {
			Float dp = dot(dRec.d, its.shFrame.n);
			if (dp > 0)
				E += directRadiance * dp;
		}

		/* Sample the indirect illumination component */
		if (handleIndirect) {
			query.newQuery(RadianceQueryRecord::ERadianceNoEmission, medium);
			Vector d = frame.toWorld(warp::squareToCosineHemisphere(query.nextSample2D()));
			++query.depth;
			query.medium = medium;
			E += Li(RayDifferential(its.p, d, its.time), query, LightField_Inf()) * M_PI;
		}

		sampler->advance();
	}

	return E / (Float)nSamples;
}

void SamplingIntegrator::cancel() {
	if (m_process)
		Scheduler::getInstance()->cancel(m_process);
}

bool SamplingIntegrator::render(Scene *scene,
	RenderQueue *queue, const RenderJob *job,
	int sceneResID, int sensorResID, int samplerResID) {
	ref<Scheduler> sched = Scheduler::getInstance();
	ref<Sensor> sensor = static_cast<Sensor *>(sched->getResource(sensorResID));
	ref<Film> film = sensor->getFilm();

	size_t nCores = sched->getCoreCount();
	const Sampler *sampler = static_cast<const Sampler *>(sched->getResource(samplerResID, 0));
	size_t sampleCount = sampler->getSampleCount();

	Log(EInfo, "Starting render job (%ix%i, " SIZE_T_FMT " %s, " SIZE_T_FMT
		" %s, " SSE_STR ") ..", film->getCropSize().x, film->getCropSize().y,
		sampleCount, sampleCount == 1 ? "sample" : "samples", nCores,
		nCores == 1 ? "core" : "cores");

	/* This is a sampling-based integrator - parallelize */
	ref<ParallelProcess> proc = new BlockedRenderProcess(job,
		queue, scene->getBlockSize());
	int integratorResID = sched->registerResource(this);
	proc->bindResource("integrator", integratorResID);
	proc->bindResource("scene", sceneResID);
	proc->bindResource("sensor", sensorResID);
	proc->bindResource("sampler", samplerResID);
	scene->bindUsedResources(proc);
	bindUsedResources(proc);
	sched->schedule(proc);

	m_process = proc;
	sched->wait(proc);
	m_process = NULL;
	sched->unregisterResource(integratorResID);

	return proc->getReturnStatus() == ParallelProcess::ESuccess;
}
Float inmiWeight(Float pdfA, Float pdfB)  {
	pdfA *= pdfA;
	pdfB *= pdfB;
	return pdfA / (pdfA + pdfB);
}
void SamplingIntegrator::bindUsedResources(ParallelProcess *) const {
	/* Do nothing by default */
}

void SamplingIntegrator::wakeup(ConfigurableObject *parent,
	std::map<std::string, SerializableObject *> &) {
	/* Do nothing by default */
}

void SamplingIntegrator::renderBlock(const Scene *scene,
	const Sensor *sensor, Sampler *sampler, ImageBlock *block,
	const bool &stop, const std::vector< TPoint2<uint8_t> > &points) {

	Float diffScaleFactor = 1.0f /
		std::sqrt((Float)sampler->getSampleCount());

	bool needsApertureSample = sensor->needsApertureSample();
	bool needsTimeSample = sensor->needsTimeSample();

	RadianceQueryRecord rRec(scene, sampler);
	Point2 apertureSample(0.5f);
	Float timeSample = 0.5f;
	RayDifferential sensorRay;

	block->clear();

	uint32_t queryType = RadianceQueryRecord::ESensorRay;

	if (!sensor->getFilm()->hasAlpha()) /* Don't compute an alpha channel if we don't have to */
		queryType &= ~RadianceQueryRecord::EOpacity;

	// TODO(yaoyi): add pdf implementation!!!!!!!!!!!!!!!!!
	// these vectors will store the light field/sampling weight/pdf information for each bounce
	std::vector<float> lf_inference_input;
	std::vector<float> sw_inference_input;
	std::vector<float> alpha_inference_input;
	// std::vector<float> pdf_inference_input;
	// NOTE(yaoyi): this part is essential because we do not know 
	// what would be the inference batch size in runtime 
	// here we assume that the input data are extremely large
	SLog(EInfo, "%d,%d,%d", lf_net_batchSize, sw_net_batchSize, alpha_net_batchSize);
	for (int i = 0; i < lf_net_batchSize * 6; ++i)
	{
		lf_inference_input.push_back(0.f);
	}
	for (int i = 0; i < alpha_net_batchSize * 6; ++i)
	{
		alpha_inference_input.push_back(0.f);
	}
	for (int i = 0; i < sw_net_batchSize * 3; ++i)
	{
		sw_inference_input.push_back(0.f);
	}
	// for (int i = 0 ; i < pdf_net_batchSize * 5 ; ++i)
	// {
	//     pdf_inference_input.push_back(0.f);
	// }
	std::vector<LightField_Inf> inf_data;//NOTE(yaoyi): importance, this vector will contain all the information for inferences and future calculation
	int lf_inference_accumulation = 0;
	int sw_inference_accumulation = 0;
	int alpha_inference_accumulation = 0;
	// int pdf_inference_accumulation = 0;

	// for tensorRT inference, init engines//TODO(yaoyi): important, enable pdf engine and pdf_stream
	std::unique_ptr<nvinfer1::IExecutionContext, nvinfer1::Destroy<nvinfer1::IExecutionContext>> lf_context{ nullptr };
	CudaStream lf_stream;
	std::unique_ptr<nvinfer1::IExecutionContext, nvinfer1::Destroy<nvinfer1::IExecutionContext>> alpha_context{ nullptr };
	CudaStream alpha_stream;
	std::unique_ptr<nvinfer1::IExecutionContext, nvinfer1::Destroy<nvinfer1::IExecutionContext>> sw_context{ nullptr };
	CudaStream sw_stream;
	// std::unique_ptr<nvinfer1::IExecutionContext, nvinfer1::Destroy<nvinfer1::IExecutionContext>> pdf_context{ nullptr };
	// CudaStream pdf_stream;

	for (size_t i = 0; i<points.size(); ++i) {
		Point2i offset = Point2i(points[i]) + Vector2i(block->getOffset());
		if (stop)
			break;



		sampler->generate(offset);
		for (size_t j = 0; j<sampler->getSampleCount(); j++) {

			// NOTE(yaoyi): this will record the lf/sw/pdf weight information for this specific sample of a specific pixel
			LightField_Inf each_sample_temp;

			rRec.newQuery(queryType, sensor->getMedium());
			Point2 samplePos(Point2(offset) + Vector2(rRec.nextSample2D()));
			if (needsApertureSample)
				apertureSample = rRec.nextSample2D();
			if (needsTimeSample)
				timeSample = rRec.nextSample1D();

			Spectrum spec = sensor->sampleRayDifferential(
				sensorRay, samplePos, apertureSample, timeSample);
			sensorRay.scaleDifferential(diffScaleFactor);

			// NOTE: this is only for one ray path
			// spec *= Li(sensorRay, rRec, LightField_Inf());

			each_sample_temp.spec = spec;//NOTE(yaoyi): this is only recorded for film writting
			each_sample_temp.alpha = rRec.alpha;//NOTE(yaoyi): this is only recorded for film writting
			each_sample_temp.samplePos = samplePos;//NOTE(yaoyi): this is only recorded for film writting
			each_sample_temp.final_li = Li(sensorRay, rRec, each_sample_temp);//NOTE(yaoyi): this will record the non-network related li values

			for (int lf_num = 0; lf_num < each_sample_temp.lf_inner.size(); lf_num++)
			{
				int total_index = lf_inference_accumulation + lf_num;
				/*if (each_sample_temp.lf_inner[lf_num].u > 0)
				{
					SLog(EInfo, "%f", each_sample_temp.lf_inner[lf_num].u);
				}*/
				if (total_index > lf_net_batchSize)
				{
					SLog(EInfo, "Why?!");
				}
				lf_inference_input[total_index * 6 + 0] = each_sample_temp.lf_inner[lf_num].posx;
				lf_inference_input[total_index * 6 + 1] = each_sample_temp.lf_inner[lf_num].posy;
				lf_inference_input[total_index * 6 + 2] = each_sample_temp.lf_inner[lf_num].posz;
				lf_inference_input[total_index * 6 + 3] = each_sample_temp.lf_inner[lf_num].dirx;
				lf_inference_input[total_index * 6 + 4] = each_sample_temp.lf_inner[lf_num].diry;
				lf_inference_input[total_index * 6 + 5] = each_sample_temp.lf_inner[lf_num].dirz;
			/*	lf_inference_input[total_index * 6 + 3] = each_sample_temp.lf_inner[lf_num].posx;
				lf_inference_input[total_index * 6 +4] = each_sample_temp.lf_inner[lf_num].posy;
				lf_inference_input[total_index * 6 + 5] = each_sample_temp.lf_inner[lf_num].posz;
				lf_inference_input[total_index * 6 + 0] = each_sample_temp.lf_inner[lf_num].dirx;
				lf_inference_input[total_index * 6 +1] = each_sample_temp.lf_inner[lf_num].diry;
				lf_inference_input[total_index * 6 + 2] = each_sample_temp.lf_inner[lf_num].dirz;*/
			}
			lf_inference_accumulation += each_sample_temp.lf_inner.size();
			for (int alpha_num = 0; alpha_num < each_sample_temp.alpha_inner.size(); alpha_num++)
			{
				int total_index = alpha_inference_accumulation + alpha_num;
				/*if (each_sample_temp.alpha_inner[alpha_num].u > 0)
				{
					SLog(EInfo, "%f", each_sample_temp.alpha_inner[alpha_num].u);
				}*/
				if (total_index > alpha_net_batchSize)
				{
					SLog(EInfo, "Why?!");
				}
				alpha_inference_input[total_index * 6 + 0] = each_sample_temp.alpha_inner[alpha_num].posx;
				alpha_inference_input[total_index * 6 + 1] = each_sample_temp.alpha_inner[alpha_num].posy;
				alpha_inference_input[total_index * 6 + 2] = each_sample_temp.alpha_inner[alpha_num].posz;
				alpha_inference_input[total_index * 6 + 3] = each_sample_temp.alpha_inner[alpha_num].dirx;
				alpha_inference_input[total_index * 6 + 4] = each_sample_temp.alpha_inner[alpha_num].diry;
				alpha_inference_input[total_index * 6 + 5] = each_sample_temp.alpha_inner[alpha_num].dirz;
				/*alpha_inference_input[total_index * 6 + 3] = each_sample_temp.alpha_inner[alpha_num].posx;
				alpha_inference_input[total_index * 6 + 4] = each_sample_temp.alpha_inner[alpha_num].posy;
				alpha_inference_input[total_index * 6 + 5] = each_sample_temp.alpha_inner[alpha_num].posz;
				alpha_inference_input[total_index * 6 + 0] = each_sample_temp.alpha_inner[alpha_num].dirx;
				alpha_inference_input[total_index * 6 + 1] = each_sample_temp.alpha_inner[alpha_num].diry;
				alpha_inference_input[total_index * 6 + 2] = each_sample_temp.alpha_inner[alpha_num].dirz;*/
			}
			alpha_inference_accumulation += each_sample_temp.alpha_inner.size();
			for (int sw_num = 0; sw_num < each_sample_temp.sw_inner.size(); sw_num++)
			{int total_index = sw_inference_accumulation + sw_num;
				if (total_index > sw_net_batchSize)
				{
					SLog(EInfo, "Why sw?!");
				}
				
				sw_inference_input[total_index * 3 + 0] = each_sample_temp.sw_inner[sw_num].theta;
				sw_inference_input[total_index * 3 + 1] = each_sample_temp.sw_inner[sw_num].phi;
				sw_inference_input[total_index * 3 + 2] = each_sample_temp.sw_inner[sw_num].radius;

			}
			sw_inference_accumulation += each_sample_temp.sw_inner.size();
			// TODO(yaoyi): pdf network data recording, need to add code here
			inf_data.push_back(each_sample_temp);// push back the recording for this sample path into the all data recording structure
			sampler->advance();
		}
	}

	std::vector<float> alpha_inference_output;
	if (alpha_inference_accumulation > 0)
	{
		void* alpha_bindings[2]{ 0 };
		if (!alpha_engine) return;
		for (int i = 0; i < alpha_engine->getNbBindings(); ++i)
		{
			Dims dims{ alpha_engine->getBindingDimensions(i) };
			size_t size = std::accumulate(dims.d + 1, dims.d + dims.nbDims, alpha_net_batchSize, multiplies<size_t>());
			cudaMalloc(&alpha_bindings[i], size * sizeof(float));

			// Resize CPU buffers to fit Tensor.
			if (alpha_engine->bindingIsInput(i)) {
				alpha_inference_input.resize(size);
			}
			else
				alpha_inference_output.resize(size);
		}
		// Create Execution Context.
		alpha_context.reset(alpha_engine->createExecutionContext());
		Dims alpha_dims_i{ alpha_engine->getBindingDimensions(0) };
		Dims2 alpha_inputDims{ alpha_net_batchSize, alpha_dims_i.d[1] };
		alpha_context->setBindingDimensions(0, alpha_inputDims);
#pragma omp critical
		{
			alpha_launchInference(alpha_context.get(), alpha_stream, alpha_inference_input, alpha_inference_output, alpha_bindings, alpha_net_batchSize);
			// Wait until the work is finished.
			cudaStreamSynchronize(alpha_stream);
		}
		for (void* ptr : alpha_bindings)
			cudaFree(ptr);
	}

	std::vector<float> lf_inference_output;
	if (lf_inference_accumulation > 0)
	{
		void* lf_bindings[2]{ 0 };
		if (!lf_engine) return;
		for (int i = 0; i < lf_engine->getNbBindings(); ++i)
		{
			Dims dims{ lf_engine->getBindingDimensions(i) };
			size_t size = std::accumulate(dims.d + 1, dims.d + dims.nbDims, lf_net_batchSize, multiplies<size_t>());
			cudaMalloc(&lf_bindings[i], size * sizeof(float));

			// Resize CPU buffers to fit Tensor.
			if (lf_engine->bindingIsInput(i)) {
				lf_inference_input.resize(size);
			}
			else
				lf_inference_output.resize(size);
		}
		// Create Execution Context.
		lf_context.reset(lf_engine->createExecutionContext());
		Dims lf_dims_i{ lf_engine->getBindingDimensions(0) };
		Dims2 lf_inputDims{ lf_net_batchSize, lf_dims_i.d[1] };
		lf_context->setBindingDimensions(0, lf_inputDims);
#pragma omp critical
		{
			lf_launchInference(lf_context.get(), lf_stream, lf_inference_input, lf_inference_output, lf_bindings, lf_net_batchSize);
			// Wait until the work is finished.
			cudaStreamSynchronize(lf_stream);
		}
		for (void* ptr : lf_bindings)
			cudaFree(ptr);
	}
	// NOTE(yaoyi): perform eval network inference 
	//	SLog(EInfo, "LF ready");
	/**************************************END***********************************************/
	// NOTE(yaoyi): sampling weight network inference ------------------------------------------------------------------
	std::vector<float> sw_inference_output;
	if (sw_inference_accumulation > 0)
	{
		void* sw_bindings[2]{ 0 };
		if (!sw_engine) return;
		for (int i = 0; i < sw_engine->getNbBindings(); ++i)
		{
			Dims dims{ sw_engine->getBindingDimensions(i) };
			size_t size = std::accumulate(dims.d + 1, dims.d + dims.nbDims, sw_net_batchSize, multiplies<size_t>());
			cudaMalloc(&sw_bindings[i], size * sizeof(float));

			// Resize CPU buffers to fit Tensor.
			if (sw_engine->bindingIsInput(i)) {
				sw_inference_input.resize(size);
			}
			else
				sw_inference_output.resize(size);
		}
		// Create Execution Context.
		sw_context.reset(sw_engine->createExecutionContext());
		Dims sw_dims_i{ sw_engine->getBindingDimensions(0) };
		Dims2 sw_inputDims{ sw_net_batchSize, sw_dims_i.d[1] };
		sw_context->setBindingDimensions(0, sw_inputDims);
		sw_launchInference(sw_context.get(), sw_stream, sw_inference_input, sw_inference_output, sw_bindings, sw_net_batchSize);
		// Wait until the work is finished.
		cudaStreamSynchronize(sw_stream);
		for (void* ptr : sw_bindings)
			cudaFree(ptr);
	}
	//	SLog(EInfo, "SW ready");
	/**********************************END***************************************************/
	//TODO(yaoyi): need to add the pdf inference here----------------------------------------------------------

	int lf_inferred_data = 0;
	int sw_inferred_data = 0;
	int alpha_inferred_data = 0;
	// TODO(yaoyi): implement pdf part!!!!!!!!!!!!!
	// int pdf_inferred_data = 0;
	// -------------------------------------Assemble the inference output-------------------------------------------------



	for (int k = 0; k < inf_data.size(); ++k) // NOTE(yaoyi): for each ray sample (total = blocksize * samplecount)
	{
		Spectrum AlphaThrough = Spectrum(1.f);
		//here if the  direct illumination
		bool ifblack = false;
		for (int i = 0; i < inf_data[k].lf_inner.size(); ++i)
		{
			if (inf_data[k].lf_inner[i].isdirect)
			{
				
				Spectrum inferred_color;
				int inferred_index = (lf_inferred_data + i);
				if (lf_inferred_data >= lf_net_batchSize)
				{
					SLog(EInfo, "Whylf?!");
				}
			//	SLog(EInfo, "infer-index%d", inferred_index);

				float res_r = lf_inference_output[inferred_index * 3 + 0];
				float res_g = lf_inference_output[inferred_index * 3 + 1];
				float res_b =lf_inference_output[inferred_index * 3 + 2];
				if (res_r < 0.1)
					res_r = 0;
				if (res_g < 0.1)
					res_g = 0;
				if (res_b < 0.1)
					res_b = 0;
				res_r = exp(std::max(res_r, 0.f))-1;
				res_g = exp(std::max(res_g, 0.f)) - 1;
				res_b = exp(std::max(res_b, 0.f)) - 1;
				inferred_color.fromLinearRGB(res_r, res_g, res_b);
				inf_data[k].final_li +=  inferred_color;
				if (inf_data[k].lf_inner[i].isblack)
				{
					inferred_color.fromLinearRGB(0, 0, 0);
			//		ifblack = true;
					break;
				}
				
			}
		}



		for (int i = 0; i < inf_data[k].alpha_inner.size(); ++i)
		{
			if (inf_data[k].alpha_inner[i].isdirect)
			{
				//SLog(EInfo, "I'm alpha");
				Spectrum inferred_color;
				int inferred_index = (alpha_inferred_data + i);
				if (alpha_inferred_data >= alpha_net_batchSize)
				{
					SLog(EInfo, "Whyalpha?!");
				}
				//	SLog(EInfo, "infer-index %d,%d", inferred_index, alpha_inference_output.size());

				float res_r = alpha_inference_output[inferred_index * 4 + 0];
				float res_g = alpha_inference_output[inferred_index * 4 + 1];
				float res_b = alpha_inference_output[inferred_index * 4 + 2];
				//SLog(EInfo, "I'm alpha1");
				res_r = std::max(res_r, 0.f);
				res_g = std::max(res_g, 0.f);
				res_b = std::max(res_b, 0.f);
				inferred_color.fromLinearRGB(res_r, res_g, res_b);
				AlphaThrough = inferred_color;
			//	SLog(EInfo, "I'm alpha2");
			}
		}


		
	//	SLog(EInfo, "lfdone");
		/*FILE *fp = fopen("D:/1.bin", "wb");*/
		for (int i = 0; i < inf_data[k].sw_inner.size(); ++i)
		{
			if (ifblack)
			{
				inf_data[k].final_li = Spectrum(0.f);
				break;
			}
				
			float swout[16 * 16 * 3];
			float pdfout[16 * 16];
			float rsum, gsum, bsum;
			rsum = 0;
			bsum = 0;
			gsum = 0;
			for (int p = 0; p < 16 * 16; p++)
			{
				int inferred_index = (sw_inferred_data + i) * 16 * 16 + p;
				if (sw_inferred_data >= sw_net_batchSize)
				{
					SLog(EInfo, "Whyswhere?!");
				}
				/*float res_r = std::max( sw_inference_output[inferred_index * 6 + 3],0.f);
				float res_g = std::max(sw_inference_output[inferred_index * 6 + 4], 0.f);
				float res_b = std::max(sw_inference_output[inferred_index * 6 + 5], 0.f);*/
				float res_r = std::max(sw_inference_output[inferred_index * 3 + 0], 0.0f);
				float res_g = std::max(sw_inference_output[inferred_index * 3 + 1], 0.0f);
				float res_b = std::max(sw_inference_output[inferred_index * 3 + 2], 0.0f);

				res_r = exp(res_r) - 1;
				rsum += res_r;
				res_g = exp(res_g) - 1;
				gsum += res_g;
				res_b = exp(res_b) - 1;
				bsum += res_b;

				swout[p * 3 + 0] = res_r;
				swout[p * 3 + 1] = res_g;
				swout[p * 3 + 2] = res_b;
				pdfout[p] = res_r + res_b + res_b;
				/*fwrite(&res_r, sizeof(int), 1, fp);
				fwrite(&res_g, sizeof(int), 1, fp);
				fwrite(&res_b, sizeof(int), 1, fp);*/
				//	fwrite(&mPhiRes, sizeof(int), 1, fp);

			}

			// START: randomly pick
			vector<float> b;
			vector<float> c;
			
			for (int p = 0; p < 16; p++)
			{
				b.push_back(pdfout[p * 16]);
				for (int q = 1; q < 16; q++)
				{
					b.push_back(b[p * 16 + q - 1] + pdfout[p * 16 + q]);

				}
				c.push_back(b[p * 16 + 16 - 1]);
				for (int q = 0; q < 16; q++)
				{
					if (c[p] != 0)
						b[p * 16 + q] /= c[p];
				}

			}
			for (int p = 1; p < 16; p++)
			{
				c[p] += c[p - 1];
			}
			for (int p = 0; p < 16; p++)
			{
				for (int q = 0; q < 16; q++)
				{
					pdfout[p * 16 + q] /= c[15];
				}
			}
			for (int p = 0; p < 16; p++)
			{
				if (c[16 - 1] != 0)
					c[p] /= c[16 - 1];
			}


			// Clamp to make sure that no random numbers contain 0 or 1.
			float ksi1 = rand() / (RAND_MAX + 1.0);
			float ksi2 = rand() / (RAND_MAX + 1.0);

			int r, z;
			r = 16 - 1;
			z = 16 - 1;
			float realr, realz;
			realr = 0;
			realz = 0;
			for (int q = 0; q < 16; q++)
			{
				if (c[q]>ksi1)
				{
					r = q;
					if (q == 0)
						realr = ksi1 / c[q] + q;
					else
						realr = (ksi1 - c[q - 1]) / (c[q] - c[q - 1]) + q;
					realr = (realr / 8.f) - 1;
					break;
				}

			}
			for (int q = 0; q < 16; q++)
			{
				if (b[r * 16 + q]>ksi2)
				{
					if (q == 0)
						realz = ksi2 / (b[r * 16 + q]) + q;

					realz = (ksi2 - b[r * 16 + q - 1]) / (b[r * 16 + q] - b[r * 16 + q - 1]) + q;
					realz = (realz / 8.f) - 1;
					z = q;
					break;
				}
			}
			
			float pdf = 0;
			pdf = pdfout[r * 16 + z];
			
//from here I will calculate the BSDF contribution


#ifdef loadlog
			SLog(EInfo, "pdf %f ", pdf);
			SLog(EInfo, "r z %d %d", r, z);
#endif
			Intersection currentIts = inf_data[k].sw_inner[i].directIts;
			Vector dir = bounding_sphere_center - currentIts.p;
			Vector nor_dir = normalize(dir);
			float randomr = dir.length() / bounding_sphere_radius;
			float solidtheta = (asin(1.0 / randomr));
			float solidradius = bounding_sphere_radius*cos(solidtheta);;
			float solidangle = 2 * M_PI*(1 - cos(solidtheta));
			float centerdist = bounding_sphere_radius*sin(solidtheta);
#ifdef loadlog

			SLog(EInfo, "theta %f ", solidtheta);


			SLog(EInfo, "solidradius %f ", solidradius);



			SLog(EInfo, "solidangle %f ", solidangle);
			SLog(EInfo, "centerdist %f ", centerdist);
#endif // log
			float centerx, centery, centerz;
			centerx = bounding_sphere_center[0] - nor_dir[0] * centerdist;
			centery = bounding_sphere_center[1] - nor_dir[1] * centerdist;
			centerz = bounding_sphere_center[2] - nor_dir[2] * centerdist;
#ifdef loadlog
			SLog(EInfo, "center %f %f %f", centerx, centery, centerz);
#endif // log
			Vector3 centerpos = Vector3(centerx, centery, centerz);
			//!!!!!!!!!!!ZJQ:this is the z-up version please change if you want a y-up version
			Vector3 strictup=Vector3(0.f,0.f,1.f);

			float upx, upy, upz;
			Vector up = dot(nor_dir, strictup)*nor_dir;
			up = strictup - up;
			up = normalize(up);
			Vector right;
			right = cross(nor_dir, up);
			Frame swlocal;
			swlocal.s = up;
			swlocal.t = right;
			swlocal.n = nor_dir;

#ifdef loadlog
			SLog(EInfo, "up %f %f %f", upx, upy, upz);
#endif // log
			float rightx, righty, rightz;
			upx = up[0];
			upy = up[1];
			upz = up[2];
			rightx = right[0];
			righty = right[1];
			rightz = right[2];
			float hitx, hity, hitz;
			hitx = centerx + (upx*realr + rightx*realz)*solidradius;
			hity = centery + (upy*realr + righty*realz)*solidradius;
			hitz = centerz + (upz*realr + rightz*realz)*solidradius;
			Point hit = Point(hitx, hity, hitz);
			Vector hitdir = normalize(hit - currentIts.p);
			float R = bounding_sphere_radius / tan(solidtheta);
			float area = (solidradius*solidradius) / 64.f;
			float disttheta = sqrt(realr*realr + realz*realz)*solidradius;
#ifdef loadlog
			SLog(EInfo, "real %f %f", realr, realz);
#endif
			float viewtocenter = dir.length() - centerdist;
			float distthetaprime = atan(disttheta / viewtocenter);
			float scalearea = 1.f;
			pdf *= 1.f / area;
			float disttoplane = sqrt(disttheta*disttheta + viewtocenter*viewtocenter);
			pdf *= disttoplane*disttoplane / cos(distthetaprime);



			float rfin, gfin, bfin;

			int findindex = r * 16 + z;
			rfin = swout[findindex * 3 + 0] / pdf;
			gfin = swout[findindex * 3 + 1] / pdf;
			bfin = swout[findindex * 3 + 2] / pdf;
			/*	rsum *= solidangle;
			gsum *= solidangle;
			bsum *= solidangle;*/
			Spectrum inferred_color;
			inferred_color.fromLinearRGB(rfin, gfin, bfin);
#ifdef loadlog
			SLog(EInfo, "final %f %f %f", rfin, gfin, bfin);

#endif 
			BSDFSamplingRecord bRec(currentIts, currentIts.toLocal(hitdir), ERadiance);

			Ray shadowRay = Ray(currentIts.p, hitdir, Epsilon, (dir.length()- bounding_sphere_radius) * (1 - ShadowEpsilon), currentIts.time);
			Spectrum tempLiFinal = Spectrum(0.f);
			if (!scene->getKDTree()->rayIntersect(shadowRay)) // inside shadow 
			{
			Spectrum bsdfVal = inf_data[k].sw_inner[i].directBSDF->eval(bRec);
			Float BSDFpdf = inf_data[k].sw_inner[i].directBSDF->pdf(bRec);
			float cosThetaVal = Frame::cosTheta(currentIts.toLocal(nor_dir));
			if (!bsdfVal.isZero() && mitsuba::dot(nor_dir, currentIts.geoFrame.n) * cosThetaVal > 0) {
			Float EMPDF = inmiWeight(pdf,BSDFpdf);
			tempLiFinal = inf_data[k].sw_inner[i].throughput * inferred_color * bsdfVal*EMPDF;
			tempLiFinal[0] = clamp(tempLiFinal[0], 0.f, 10.f);
			tempLiFinal[1] = clamp(tempLiFinal[1], 0.f, 10.f);
			tempLiFinal[2] = clamp(tempLiFinal[2], 0.f, 10.f);
			}
					}
			//JUNQIU:here
			//	inf_data[k].final_li += tempLiFinal*Alpha;//NOTE(yaoyi): add the li value of the sampling weight network to the final li result
			inf_data[k].final_li += tempLiFinal*AlphaThrough;
			//HERE The BSDF sampling part
			for (int j = 0; j < inf_data[k].lf_inner.size(); j++)
			{
				//SLog(EInfo, "index %d", inf_data[k].lf_inner[i].collectindex);
				if (inf_data[k].lf_inner[j].collectindex == i)
				{
					int inferred_index = lf_inferred_data + j;
					Spectrum inferred_color;
					/*float res_r = exp(lf_inference_output[inferred_index * 3 + 0]) - 1.f;
					float res_g = exp(lf_inference_output[inferred_index * 3 + 1]) - 1.f;
					float res_b = exp(lf_inference_output[inferred_index * 3 + 2]) - 1.f;*/
					float res_r = exp(lf_inference_output[inferred_index * 3 + 0])-1;
					float res_g = exp(lf_inference_output[inferred_index * 3 + 1])-1;
					float res_b = exp(lf_inference_output[inferred_index * 3 + 2])-1;
					res_r = std::max(res_r, 0.f);
					res_g = std::max(res_g, 0.f);
					res_b = std::max(res_b, 0.f);
					inferred_color.fromLinearRGB(res_r, res_g, res_b);
				//	SLog(EInfo, "res_r%f", res_r);
				/*	if (res_r > 0)
					{
						SLog(EInfo, "22222");
					}*/
					Vector dirintersect;
				/*	Float intersectu = inf_data[k].lf_inner[j].u;
					Float intersectv = inf_data[k].lf_inner[j].v;*/
					Point intersect_bounding;
					intersect_bounding.z = inf_data[k].lf_inner[j].posz;
					float jiaozheng = sqrt(1 - intersect_bounding.z*intersect_bounding.z);
					//float angle = intersectv*M_PI * 2;
					intersect_bounding.x = inf_data[k].lf_inner[j].posx;
					intersect_bounding.y = inf_data[k].lf_inner[j].posy;
		//			intersect_bounding = normalize(intersect_bounding);
					Vector in_dir = intersect_bounding - currentIts.p;
					Vector in_nor_dir = normalize(dir);
					Vector in_loacaldir = swlocal.toLocal(in_nor_dir);
					Float len_scale = 1 / dot(in_nor_dir, nor_dir);
					Float up_len = in_loacaldir.x*len_scale;
					Float right_len = in_loacaldir.y*len_scale;
					Float far_len = tan(solidtheta);
					int upindex = (up_len / far_len) * 8 + 8;
					int rightindex= (right_len / far_len) * 8 + 8;
					if ((upindex * 16 + rightindex) < 0 || (upindex * 16 + rightindex) >= 16 * 16)
					{
						upindex = 15;
						rightindex = 15;
					}
					float B_pdf = pdfout[upindex * 16 + rightindex];
				//	SLog(EInfo, "B_pdf%f", B_pdf);
					B_pdf *= 1.f / area;
				//	SLog(EInfo, "B_pdf%f", B_pdf);
					float B_disttheta= sqrt((up_len/far_len)*(up_len / far_len) + (right_len / far_len)*(right_len / far_len))*solidradius;
					float B_disttoplane = sqrt(B_disttheta*B_disttheta + viewtocenter*viewtocenter);
					B_pdf *= B_disttoplane*B_disttoplane / dot(in_nor_dir, nor_dir);
				//	SLog(EInfo, "B_pdf%f", B_pdf);
					float BSDF_PDF= inf_data[k].lf_inner[j].bsdfpdf;
					float MI = inmiWeight(BSDF_PDF, B_pdf);
				//	SLog(EInfo, "res_r%f", MI);
					Spectrum tempLiSW = inf_data[k].lf_inner[j].throughput * inferred_color*MI*AlphaThrough;
					tempLiSW[0] = clamp(tempLiSW[0], 0.f, 10.f);
					tempLiSW[1] = clamp(tempLiSW[1], 0.f, 10.f);
					tempLiSW[2] = clamp(tempLiSW[2], 0.f, 10.f);
					inf_data[k].final_li += tempLiSW;
				}
			}

		}
	//	SLog(EInfo, "swdone");
		sw_inferred_data += inf_data[k].sw_inner.size();
		lf_inferred_data+= inf_data[k].lf_inner.size();
		alpha_inferred_data+= inf_data[k].alpha_inner.size();
		// TODO(yaoyi): need to add the values of the pdf network, but it is possible that it should be inferred before the eval and sampling weight network
		Spectrum spec_final = inf_data[k].final_li;
		spec_final *= inf_data[k].spec;//NOTE(yaoyi): this is similar to the way where originally the li should be treat in the renderBlock function
		block->put(inf_data[k].samplePos, spec_final, inf_data[k].alpha);//NOTE(yaoyi): write everything above into the film
	}
}

MonteCarloIntegrator::MonteCarloIntegrator(const Properties &props) : SamplingIntegrator(props) {
	/* Depth to begin using russian roulette */
	m_rrDepth = props.getInteger("rrDepth", 5);

	/* Longest visualized path depth (\c -1 = infinite).
	A value of \c 1 will visualize only directly visible light sources.
	\c 2 will lead to single-bounce (direct-only) illumination, and so on. */
	m_maxDepth = props.getInteger("maxDepth", -1);

	/**
	* This parameter specifies the action to be taken when the geometric
	* and shading normals of a surface don't agree on whether a ray is on
	* the front or back-side of a surface.
	*
	* When \c strictNormals is set to \c false, the shading normal has
	* precedence, and rendering proceeds normally at the risk of
	* introducing small light leaks (this is the default).
	*
	* When \c strictNormals is set to \c true, the random walk is
	* terminated when encountering such a situation. This may
	* lead to silhouette darkening on badly tesselated meshes.
	*/
	m_strictNormals = props.getBoolean("strictNormals", false);

	/**
	* When this flag is set to true, contributions from directly
	* visible emitters will not be included in the rendered image
	*/
	m_hideEmitters = props.getBoolean("hideEmitters", false);

	if (m_rrDepth <= 0)
		Log(EError, "'rrDepth' must be set to a value greater than zero!");

	if (m_maxDepth <= 0 && m_maxDepth != -1)
		Log(EError, "'maxDepth' must be set to -1 (infinite) or a value greater than zero!");
}

MonteCarloIntegrator::MonteCarloIntegrator(Stream *stream, InstanceManager *manager)
	: SamplingIntegrator(stream, manager) {
	m_rrDepth = stream->readInt();
	m_maxDepth = stream->readInt();
	m_strictNormals = stream->readBool();
	m_hideEmitters = stream->readBool();
}

void MonteCarloIntegrator::serialize(Stream *stream, InstanceManager *manager) const {
	SamplingIntegrator::serialize(stream, manager);
	stream->writeInt(m_rrDepth);
	stream->writeInt(m_maxDepth);
	stream->writeBool(m_strictNormals);
	stream->writeBool(m_hideEmitters);
}

std::string RadianceQueryRecord::toString() const {
	std::ostringstream oss;
	oss << "RadianceQueryRecord[" << endl
		<< "  type = { ";
	if (type & EEmittedRadiance) oss << "emitted ";
	if (type & ESubsurfaceRadiance) oss << "subsurface ";
	if (type & EDirectSurfaceRadiance) oss << "direct ";
	if (type & EIndirectSurfaceRadiance) oss << "indirect ";
	if (type & ECausticRadiance) oss << "caustic ";
	if (type & EDirectMediumRadiance) oss << "inscatteredDirect ";
	if (type & EIndirectMediumRadiance) oss << "inscatteredIndirect ";
	if (type & EDistance) oss << "distance ";
	if (type & EOpacity) oss << "opacity ";
	if (type & EIntersection) oss << "intersection ";
	oss << "}," << endl
		<< "  depth = " << depth << "," << endl
		<< "  its = " << indent(its.toString()) << endl
		<< "  alpha = " << alpha << "," << endl
		<< "  extra = " << extra << "," << endl
		<< "]" << endl;
	return oss.str();
}


MTS_IMPLEMENT_CLASS(Integrator, true, NetworkedObject)
MTS_IMPLEMENT_CLASS(SamplingIntegrator, true, Integrator)
MTS_IMPLEMENT_CLASS(MonteCarloIntegrator, true, SamplingIntegrator)
MTS_NAMESPACE_END
